@@ -2,7 +2,6 @@ __copyright__ = "Copyright (C) 2014-2017  Martin Blais"
 __license__ = "GNU GPLv2"
 
 import datetime
-import io
 import unittest
 import textwrap
 
@@ -14,15 +13,16 @@ from beancount.core.number import D
 from beancount.core import inventory
 from beancount.core.inventory import from_string as I
 from beancount.parser import cmptest
+from beancount.utils.test_utils import docfile
 from beancount import loader
 
 import beanquery
 
 from beanquery import CompilationError
 from beanquery import query_compile as qc
-from beanquery import query_env as qe
 from beanquery import query_execute as qx
 from beanquery import tables
+from beanquery import compiler
 
 
 class QueryBase(cmptest.TestCase):
@@ -37,7 +37,7 @@ class QueryBase(cmptest.TestCase):
         return self.ctx.compile(self.ctx.parse(query))
 
     def check_query(self, input_string, query, expected_types, expected_rows):
-        entries, errors, options = loader.load_string(input_string)
+        entries, errors, options = loader.load_string(textwrap.dedent(input_string))
         ctx = beanquery.connect('beancount:', entries=entries, errors=errors, options=options)
         curs = ctx.execute(query)
         self.assertEqual(tuple(expected_types), curs.description)
@@ -201,6 +201,7 @@ class TestFundamentals(QueryBase):
         self.assertError ("SELECT 1970-01-01 + 2022-04-01")
         self.assertResult("SELECT 2022-04-01 + 1", datetime.date(2022, 4, 2))
         self.assertResult("SELECT 1 + 2022-04-01", datetime.date(2022, 4, 2))
+        self.assertResult("SELECT 'ab' + 'c'", 'abc')
 
         # sub
         self.assertResult("SELECT 1 - 1", 0)
@@ -239,6 +240,10 @@ class TestFundamentals(QueryBase):
         self.assertResult("SELECT 'foobarbaz' !~ 'bar'", False)
         self.assertResult("SELECT 'foobarbaz' !~ 'quz'", True)
 
+        # matches
+        self.assertResult("SELECT '[0-9]+' ?~ '123'", True)
+        self.assertResult("SELECT '[0-9]+' ?~ 'ABC'", False)
+
         # and
         self.assertResult("SELECT 1 and FALSE", False)
         self.assertResult("SELECT 'something' and FALSE", False)
@@ -401,6 +406,14 @@ class TestFundamentals(QueryBase):
         self.assertResult('''SELECT interval("1 baz")''', None, relativedelta)
         self.assertResult('''SELECT interval("A days")''', None, relativedelta)
 
+    def test_date_bin(self):
+        self.assertResult('''SELECT date_bin(interval('1 year'), 2024-11-10, 2024-06-01)''', datetime.date(2024, 6, 1))
+        self.assertResult('''SELECT date_bin('1 year', 2024-11-10, 2024-06-01)''', datetime.date(2024, 6, 1))
+        self.assertResult('''SELECT date_bin('1 year', 2024-11-10, 2025-06-01)''', datetime.date(2024, 6, 1))
+        self.assertResult('''SELECT date_bin('1 month', 2024-11-10, 2024-06-03)''', datetime.date(2024, 11, 3))
+        self.assertResult('''SELECT date_bin('3 days', 2024-11-10, 2024-11-02)''', datetime.date(2024, 11, 8))
+        self.assertResult('''SELECT date_bin('3 days', 2024-11-10, 2024-11-14)''', datetime.date(2024, 11, 8))
+
 
 class TestBeancountFunctions(QueryBase):
     INPUT = textwrap.dedent("""
@@ -490,14 +503,18 @@ class TestBeancountMetadataFunctions(QueryBase):
 
 class TestFilterEntries(CommonInputBase, QueryBase):
 
+    def compile(self, query):
+        # use the ``entries`` table as default table for queries
+        c = compiler.Compiler(self.ctx)
+        c.table = self.ctx.tables.get('entries')
+        return c.compile(self.ctx.parse(query))
+
     @staticmethod
     def filter_entries(query):
         entries = []
         expr = query.c_where
-        context = qe.Row(entries, query.table.options)
-        for entry in query.table.prepare():
-            context.entry = entry
-            if expr is None or expr(context):
+        for entry in query.table:
+            if expr is None or expr(entry):
                 entries.append(entry)
         return entries
 
@@ -663,26 +680,6 @@ class TestFilterEntries(CommonInputBase, QueryBase):
         """), filtered_entries)
 
 
-class TestExecutePrint(CommonInputBase, QueryBase):
-
-    def test_print_with_filter(self):
-        oss = io.StringIO()
-        qx.execute_print(self.compile("PRINT FROM year = 2012"), oss)
-
-        self.assertEqualEntries("""
-
-          2012-02-02 * "Dinner with Dos"
-            Assets:Bank:Checking                                                   102.00 USD
-            Expenses:Restaurant                                                   -102.00 USD
-
-        """, oss.getvalue())
-
-    def test_print_with_no_filter(self):
-        oss = io.StringIO()
-        qx.execute_print(self.compile("PRINT"), oss)
-        self.assertEqualEntries(self.INPUT, oss.getvalue())
-
-
 class TestAllocation(unittest.TestCase):
 
     def test_allocator(self):
@@ -1555,7 +1552,7 @@ class TestExecuteTables(QueryBase):
 
     def execute(self, query):
         query = self.compile(query)
-        return qx.execute_query(query)
+        return query()
 
     def test_null_table(self):
         self.assertEqual(
@@ -1619,3 +1616,268 @@ class TestInSubquery(QueryBase):
             ('Income:Contributions:Three', datetime.date(2024, 10, 28)),
             ]
         )
+
+
+class TestArrayOps(QueryBase):
+    data = """
+      1970-01-01 open Assets:Cash
+      2000-01-01 open Expenses:One
+      2000-01-02 open Expenses:Two
+      2000-01-03 open Expenses:Three
+
+      2025-01-01 * "One"
+        Expenses:One                            100.00 USD
+        Assets:Cash
+
+      2025-01-02 * "Two"
+        Expenses:Two                             10.00 USD
+        Assets:Cash
+
+      2025-01-03 * "Three"
+        Expenses:Three                           50.00 USD
+        Assets:Cash
+
+      2025-01-03 * "Three"
+        Expenses:Three                            1.00 USD
+        Assets:Cash
+    """
+
+    def test_in_accounts_entries(self):
+        self.check_query(self.data, """
+            SELECT date
+            FROM #entries
+            WHERE 'Expenses:Two' IN accounts
+            """,
+            (('date', datetime.date),),
+            [
+            (datetime.date(2000, 1, 2),),
+            (datetime.date(2025, 1, 2),),
+            ]
+        )
+
+    def test_in_accounts_transactions(self):
+        self.check_query(self.data, """
+            SELECT date, narration
+            FROM #transactions
+            WHERE 'Expenses:Two' IN accounts
+            """,
+            (('date', datetime.date), ('narration', str)),
+            [
+            (datetime.date(2025, 1, 2), 'Two'),
+            ]
+        )
+
+    def test_mathes_any_accounts_transactions(self):
+        self.check_query(self.data, """
+            SELECT date, narration
+            FROM #transactions
+            WHERE ':Two' ?~ ANY(accounts)
+            """,
+            (('date', datetime.date), ('narration', str)),
+            [
+            (datetime.date(2025, 1, 2), 'Two'),
+            ]
+        )
+
+    def test_mathes_all_accounts_transactions(self):
+        self.check_query(self.data, """
+            SELECT date, narration
+            FROM #transactions
+            WHERE '(?i):two|:cash' ?~ ALL(accounts)
+            """,
+            (('date', datetime.date), ('narration', str)),
+            [
+            (datetime.date(2025, 1, 2), 'Two'),
+            ]
+        )
+
+
+class TestArrayOpsSubquery(unittest.TestCase):
+    @classmethod
+    def setUpClass(cls):
+        cls.conn = beanquery.connect('test:test?end=10')
+
+    def check_query(self, query, expected_types, expected_rows):
+        curs = self.conn.execute(query)
+        self.assertEqual(tuple(expected_types), curs.description)
+        result_rows = curs.fetchall()
+        self.assertEqual(expected_rows, result_rows)
+
+    def test_in_subquery(self):
+        self.check_query(
+            """
+            SELECT 1 IN (SELECT x FROM #test) AS y FROM #
+            """,
+            (('y', bool),),
+            [
+            (True,),
+            ]
+        )
+        self.check_query(
+            """
+            SELECT 1 IN (SELECT x FROM #test WHERE x % 2 = 0) AS y FROM #
+            """,
+            (('y', bool),),
+            [
+            (False,),
+            ]
+        )
+
+    def test_any_subquery(self):
+        self.check_query(
+            """
+            SELECT 1 = ANY (SELECT x FROM #test) AS y FROM #
+            """,
+            (('y', bool),),
+            [
+            (True,),
+            ]
+        )
+        self.check_query(
+            """
+            SELECT 1 = ANY (SELECT x FROM #test WHERE x % 2 = 0) AS y FROM #
+            """,
+            (('y', bool),),
+            [
+            (False,),
+            ]
+        )
+
+    def test_all_subquery(self):
+        self.check_query(
+            """
+            SELECT 1 = ALL (SELECT x FROM #test) AS y FROM #
+            """,
+            (('y', bool),),
+            [
+            (False,),
+            ]
+        )
+        self.check_query(
+            """
+            SELECT 1 = ALL (SELECT x FROM #test WHERE x = 1) AS y FROM #
+            """,
+            (('y', bool),),
+            [
+            (True,),
+            ]
+        )
+
+
+class TestCreateTable(unittest.TestCase):
+
+    def setUp(cls):
+        cls.conn = beanquery.connect('')
+
+    def test_create_table(self):
+        curs = self.conn.execute('''CREATE TABLE abcd (a int, b bool, c str, d date)''')
+        r = curs.fetchall()
+        self.assertEqual(r, [])
+        names = list(self.conn.tables['abcd'].columns.keys())
+        self.assertEqual(names, ['a', 'b', 'c', 'd'])
+        types = [column.dtype for column in self.conn.tables['abcd'].columns.values()]
+        self.assertEqual(types, [int, bool, str, datetime.date])
+
+    def test_create_table_as(self):
+        self.conn.execute('''CREATE TABLE abc (a int, b bool, c str)''')
+        self.conn.execute('''INSERT INTO abc (a, b, c) VALUES (1, true, '2025/01/01')''')
+        curs = self.conn.execute('''CREATE TABLE test AS SELECT a, parse_date(c, '%Y/%m/%d') AS date FROM abc''')
+        self.assertEqual(curs.fetchall(), [])
+        names = list(self.conn.tables['test'].columns.keys())
+        self.assertEqual(names, ['a', 'date'])
+        types = [column.dtype for column in self.conn.tables['test'].columns.values()]
+        self.assertEqual(types, [int, datetime.date])
+        curs = self.conn.execute('''SELECT * FROM test''')
+        self.assertEqual(curs.fetchall(), [(1, datetime.date(2025, 1, 1))])
+
+
+class TestInsert(unittest.TestCase):
+
+    def setUp(self):
+        self.conn = beanquery.connect('')
+        self.conn.execute('''CREATE TABLE abcd (a int, b bool, c str, d date)''')
+
+    def test_insert_values(self):
+        curs = self.conn.execute('''INSERT INTO abcd (a, b, c, d) VALUES (1, TRUE, 'one', 2025-01-01)''')
+        self.assertEqual(self.conn.tables['abcd'].data[0], (1, True, 'one', datetime.date(2025, 1, 1)))
+        self.assertEqual(curs.fetchall(), [])
+
+    def test_insert_incomplete_row(self):
+        curs = self.conn.execute('''INSERT INTO abcd (a, d) VALUES (1, 2025-01-01)''')
+        self.assertEqual(self.conn.tables['abcd'].data[0], (1, None, None, datetime.date(2025, 1, 1)))
+        self.assertEqual(curs.fetchall(), [])
+
+    def test_insert_order(self):
+        curs = self.conn.execute('''INSERT INTO abcd (a, d, c, b) VALUES (1, 2025-01-01, 'one', FALSE)''')
+        self.assertEqual(self.conn.tables['abcd'].data[0], (1, False, 'one', datetime.date(2025, 1, 1)))
+        self.assertEqual(curs.fetchall(), [])
+
+    def test_insert_placeholders(self):
+        values = (1, True, 'one', datetime.date(2025, 1, 1))
+        curs = self.conn.execute('''INSERT INTO abcd (a, b, c, d) VALUES (%s, %s, %s, %s)''', values)
+        self.assertEqual(self.conn.tables['abcd'].data[0], values)
+        self.assertEqual(curs.fetchall(), [])
+
+
+class TestCSVTable(unittest.TestCase):
+
+    def setUp(self):
+        self.conn = beanquery.connect('')
+
+    @docfile
+    def test_create_table(self, filename):
+        '''\
+        1234, 1, baz, 2025-01-01
+        5678, 0, qux, 2025-01-02
+        '''
+        using = f'csv:{filename}'
+        curs = self.conn.execute(f'''CREATE TABLE test (a int, b bool, c str, d date) USING {using!r}''')
+        self.assertEqual(curs.fetchall(), [])
+        curs = self.conn.execute('''SELECT * FROM test''')
+        self.assertEqual(curs.fetchone(), (1234, True, 'baz', datetime.date(2025, 1, 1)))
+        self.assertEqual(curs.fetchone(), (5678, False, 'qux', datetime.date(2025, 1, 2)))
+        curs = self.conn.execute('''SELECT * FROM test''')
+        self.assertEqual(len(curs.fetchall()), 2)
+
+    @docfile
+    def test_create_table_header(self, filename):
+        '''\
+        id
+        1234
+        5678
+        '''
+        using = f'csv:{filename}?header=1'
+        curs = self.conn.execute(f'''CREATE TABLE test (a int) USING {using!r}''')
+        self.assertEqual(curs.fetchall(), [])
+        curs = self.conn.execute('''SELECT * FROM test''')
+        self.assertEqual(len(curs.fetchall()), 2)
+
+    @docfile
+    def test_create_table_guess_types(self, filename):
+        '''\
+        id, name, check, date, value
+        1234, one, true, 2025-01-01, 1.234
+        5678, two, false, 2025-01-02, 5.678
+        '''
+        using = f'csv:{filename}'
+        curs = self.conn.execute(f'''CREATE TABLE test USING {using!r}''')
+        self.assertEqual(curs.fetchall(), [])
+        names = list(self.conn.tables['test'].columns.keys())
+        self.assertEqual(names, ['id', 'name', 'check', 'date', 'value'])
+        types = [column.dtype for column in self.conn.tables['test'].columns.values()]
+        self.assertEqual(types, [int, str, bool, datetime.date, Decimal])
+
+
+class TestCSVSource(unittest.TestCase):
+
+    @docfile
+    def test_csv_source(self, filename):
+        '''\
+        id, name, check, date, value
+        1234, one, true, 2025-01-01, 1.234
+        '''
+        conn = beanquery.connect(f'csv:{filename}?name=test')
+        names = list(conn.tables['test'].columns.keys())
+        self.assertEqual(names, ['id', 'name', 'check', 'date', 'value'])
+        types = [column.dtype for column in conn.tables['test'].columns.values()]
+        self.assertEqual(types, [int, str, bool, datetime.date, Decimal])
